package org.fictus

import java.lang.reflect.Proxy
import java.lang.reflect.Method
import scala.collection.mutable.ListBuffer
import org.objenesis.ObjenesisStd
import net.sf.cglib.proxy._

class WtfError(msg: String) extends Error(msg)
class ExpectationError(msg: String) extends Error(msg)
class PebkacError(msg: String) extends Error(msg)
class VerificationError(msg: String) extends Error(msg)
class MatcherError(msg: String) extends Error(msg)

class ExceptionWrapper(e: Exception) extends RuntimeException(e)

trait Behavior {
    def handle: Any
}
case class Returns(returnValue: Any) extends Behavior {
    override def handle = returnValue
}
case class Throws(e: Exception) extends Behavior {
    override def handle = throw new ExceptionWrapper(e)
}
case class Forever(behavior: Behavior) extends Behavior {
    override def handle = behavior.handle
}
case class Nop() extends Behavior {
    override def handle = ()
}

case class RecordedMethod(mock: Mock, method: Method, args: List[Any]) {
    def toRecord(matchers: List[Matcher]): Record = {
        val parameterTypes = method.getParameterTypes.toList

        if (matchers.size > parameterTypes.size)
            throw new MatcherError("You have specified more matchers than arguments... only one matcher per argument will work.")
        if ((!args.isEmpty && !matchers.isEmpty) && (matchers.size != parameterTypes.size))
            throw new MatcherError("If you provide one matcher, you must provide matchers for all of the arguments!")

        val validMatchers = matchers match {
            case Nil => args.map(Equal(_))
            case _   => matchers
        }
        Record(mock, method, validMatchers)
    }
}
case class RecordedBehavior(mock: Mock, method: Method, matchers: List[Matcher], behavior: Option[Behavior] = None) {
    val methodName = mock + "." + method.getName + matchers.mkString("(", ", ", ")")
}
case class Record(mock: Mock, method: Method, matchers: List[Matcher], behaviors: ListBuffer[Behavior] = new ListBuffer()) {
    def returns(arg: Any) = {
        behaviors.append(Returns(arg))
        this
    }
    def throws(e: Exception) = {
        behaviors.append(Throws(e))
        this
    }

    def -->(arg: Any) = returns(arg)
    def -->(e: Exception) = throws(e)

    def anyTimes = {
        method.getReturnType == classOf[Unit] match {
            case true => behaviors.append(Forever(Nop()))
            case false => behaviors.lastOption match {
                case Some(last) => behaviors.append(Forever(last))
                case None       => throw new PebkacError("FOREVER WHAT!?")
            }
        }
        this
    }

    val methodName = mock + "." + method.getName + matchers.mkString("(", ", ", ")")
}

object Mock {
    def createControl() = new Control()

    def mocks2[M1](implicit control: Control, m1: Manifest[M1]): (M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"))
    }

    def mocks3[M1](implicit control: Control, m1: Manifest[M1]): (M1, M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"), mock(nameBase + "._3"))
    }

    def mock[M1, M2](implicit control: Control, m1: Manifest[M1], m2: Manifest[M2]): (M1, M2) = (mock[M1], mock[M2])
    def mock[M1, M2, M3](implicit control: Control, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3]): (M1, M2, M3) = (mock[M1], mock[M2], mock[M3])
    def mock[M1, M2, M3, M4](implicit control: Control, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3], m4: Manifest[M4]): (M1, M2, M3, M4) = (mock[M1], mock[M2], mock[M3], mock[M4])

    /** @return a mock of the given type with a name of the given type. */
    def mock[A](implicit control: Control, manifest: Manifest[A]): A = mock(manifest.erasure.getSimpleName)

    /** @return a mock of the given type with the passed in name. */
    def mock[A](name: String)(implicit control: Control, manifest: Manifest[A]): A = {
        val cl = getClass.getClassLoader

        def createProxyClass(mockedType: Class[_]) = {
            val enhancer = new Enhancer() {
                override def filterConstructors(sc: Class[_], constructors: java.util.List[_]) {}
            }

            enhancer.setClassLoader(cl)
            enhancer.setUseFactory(true)
            enhancer.setSuperclass(mockedType)
            enhancer.setCallbackTypes(Array(classOf[net.sf.cglib.proxy.InvocationHandler], classOf[NoOp]));
            enhancer.setCallbackFilter(new CallbackFilter() {
                override def accept(method: Method): Int = if (method.isBridge) 1 else 0
            })
            enhancer.createClass()
        }

        val mock = new Mock(name)
        val mockedClass = manifest.erasure
        mockedClass.isInterface match {
            case true => Proxy.newProxyInstance(cl, Array(mockedClass), mock.interfaceInvoke).asInstanceOf[A]
            case false =>
                val proxyClass = createProxyClass(mockedClass)
                val factory = Mock.objenesis.newInstance(proxyClass).asInstanceOf[Factory]
                factory.setCallbacks(Array(mock.classInvoke, NoOp.INSTANCE))
                factory.asInstanceOf[A]
        }
    }

    implicit def expect[A](expected: => A)(implicit control: Control): Record = {
        expected
        val matchers = control.matchers.get
        val record = control.last.toRecord(matchers)
        control.records.append(record)
        record
    }

    def test[A](unitFunction: => A)(implicit control: Control): A = {
        val expectations = RecordedExpectations(control.records.toList)
        control.expectations = expectations

        // set mocks to replay state
        control.isRecording = false

        try {
            val result = unitFunction
            expectations.verifyAllExpectationsAreConsumed
            result
        }
        catch {
            case e: ExceptionWrapper => throw e.getCause
        }
    }

    val objenesis = new ObjenesisStd
}

class Mock(name: String)(implicit control: Control) {

    override val toString = name

    def interfaceInvoke = new java.lang.reflect.InvocationHandler() {
        override def invoke(proxy: Any, method: Method, argsShit: Array[AnyRef]) = mockInvoke(proxy, method, argsShit)
    }

    def classInvoke = new net.sf.cglib.proxy.InvocationHandler() {
        override def invoke(receiver: AnyRef, method: Method, argsShit: Array[AnyRef]): AnyRef = mockInvoke(receiver, method, argsShit)
    }

    private def mockInvoke(proxy: Any, method: Method, argsShit: Array[AnyRef]): AnyRef = {
        val methodName = method.getName
        methodName match {
            case "toString" => "Mocked " + name
            case "equals" => java.lang.Boolean.valueOf(true)
            case "hashCode" => java.lang.Integer.valueOf(1)
            case _ =>
                val args = Option(argsShit) match {
                    case Some(args) => args.toList
                    case None       => Nil
                }

                control.isRecording match {
                    case true  => record(method, args)
                    case false => replay(method, args)
                }
        }
    }

    private def replay(method: Method, actualArgs: List[AnyRef]): AnyRef = {
        control.expectations.remove(this, method) match {
            case Some(expectation) =>
                val matchers = expectation.matchers

                val argsMatch = matchers.zip(actualArgs).map { tuple =>
                    val matcher = tuple._1
                    val actual = tuple._2
                    matcher.matches(actual)
                }

                if (!argsMatch.isEmpty && argsMatch.forall(_ == false)) {
                    def markErrors(list: List[_]) = {
                        argsMatch.zip(list).map { tuple =>
                            tuple._1 match {
                                case true  => tuple._2.toString
                                case false => "**" + tuple._2 + "**"
                            }
                        }
                    }

                    val msg = "oh shit!\n" +
                        "expected: " + name + "." + method.getName + markErrors(matchers).mkString("(", ", ", ")") + "\n" +
                        "actual:   " + name + "." + method.getName + markErrors(actualArgs).mkString("(", ", ", ")")
                    throw new ExpectationError(msg)
                }

                expectation.behavior match {
                    case None => Unit //
                    case Some(behavior) => behavior.handle match {
                        case ref: AnyRef => ref
                        case value => value match {
                            case i: Boolean => java.lang.Boolean.valueOf(i)
                            case i: Byte    => java.lang.Byte.valueOf(i)
                            case i: Char    => java.lang.Character.valueOf(i)
                            case i: Short   => java.lang.Short.valueOf(i)
                            case i: Int     => java.lang.Integer.valueOf(i)
                            case i: Long    => java.lang.Long.valueOf(i)
                            case i: Float   => java.lang.Float.valueOf(i)
                            case i: Double  => java.lang.Double.valueOf(i)
                        }
                    }
                }
            case None => throw new PebkacError("Unexpected method called: " + name + "." + method.getName + actualArgs.mkString("(", ", ", ")"))
        }
    }

    private def record(method: Method, args: List[AnyRef]): AnyRef = {
        control.last = RecordedMethod(this, method, args)
        Types.defaultValue(method.getReturnType)
    }
}

private[fictus] object RecordedExpectations {
    private def checkForErrors(records: List[Record]) {
        records.toList.filterNot(returnsUnit).filter(_.behaviors.isEmpty) map (_.methodName) match {
            case Nil =>
            case errors => throw new VerificationError("\nExpected methods are missing behaviors (returns or throws):\n" +
                    errors.mkString("", "\n", "\n") +
                    "E.g. expect(" + errors.head + ").returns(<some return>) OR\n" +
                    "     expect(" + errors.head + ").throws(<some exception>)")
        }
        records.toList.filter(returnsUnit).filter(!_.behaviors.isEmpty).filter(invalidUnitBehaviors)  map (_.methodName)match {
            case Nil =>
            case errors => throw new VerificationError("\nExpected methods have invalid behaviors (returns):\n" +
                    errors.map(_.toString + " has a return type of Unit").mkString("", "\n", "\n"))
        }
    }

    def apply(records: List[Record]) = {
        checkForErrors(records)
        val expectations = records.flatMap { record =>
            val matchers = record.matchers
            val mock = record.mock
            val method = record.method

            returnsUnit(record) && record.behaviors.isEmpty match {
                case true => List(RecordedBehavior(mock, method, matchers, None))
                case false => record.behaviors.toList.map { behavior =>
                    RecordedBehavior(mock, method, matchers, Some(behavior))
                }
            }
        }
        new RecordedExpectations(ListBuffer() ++ expectations)
    }

    private def returnsUnit(record: Record) = record.method.getReturnType == classOf[Unit]
    private def invalidUnitBehaviors(record: Record) = record.behaviors.forall(_.isInstanceOf[Returns])
}

private[fictus] class RecordedExpectations(expectations: ListBuffer[RecordedBehavior]) {
    def remove(mock: Mock, method: Method): Option[RecordedBehavior] = {
        expectations.zipWithIndex
            .find(expectation => (mock eq expectation._1.mock) && method == expectation._1.method)
            .map(expectation =>
                expectation._1.behavior match {
                    case Some(behavior) =>
                        if (behavior.isInstanceOf[Forever]) expectation._1
                        else expectations.remove(expectation._2)
                    case None => expectations.remove(expectation._2)
                })
    }

    def verifyAllExpectationsAreConsumed {
        val regularExpections = expectations.toList.filterNot(_.behavior.getOrElse(false).isInstanceOf[Forever])
        if (!regularExpections.isEmpty)
            throw new VerificationError("Expected methods were not invoked by the unit!\n" + regularExpections.map(_.methodName).mkString("\n"))
    }

    override def toString = getClass.getSimpleName + "(" + expectations + ")"
}

object Types {
    def defaultValue[A](clazz: Class[A]) =
        if (clazz == classOf[Unit]) null // NB In java, functions that return void actually return null :D
        else if (clazz == classOf[String]) "default-string"
        else if (clazz == classOf[Boolean]) java.lang.Boolean.FALSE
        else if (clazz == classOf[Byte]) java.lang.Byte.valueOf("0")
        else if (clazz == classOf[Short]) java.lang.Short.valueOf("0")
        else if (clazz == classOf[Char]) java.lang.Character.valueOf('0')
        else if (clazz == classOf[Int]) java.lang.Integer.valueOf("0")
        else if (clazz == classOf[Long]) java.lang.Long.valueOf("0")
        else if (clazz == classOf[Float]) java.lang.Float.valueOf("0")
        else if (clazz == classOf[Double]) java.lang.Double.valueOf("0")
        else null
}

class Control {

    // TODO :'( var
    var isRecording = true
    var last: RecordedMethod = _
    var expectations: RecordedExpectations = _

    val records = new ListBuffer[Record]

    val matchers = new RecordedMatchers(this)
}

private[fictus] class RecordedMatchers(control: Control) {
    private val matchers = new ListBuffer[Matcher]

    def append(matcher: Matcher) {
        if (!control.isRecording) throw new WtfError("You are using a matcher in the replay state! This won't work...")
        matchers.append(matcher)
    }

    def get: List[Matcher] = {
        val theMatchers = matchers.toList
        matchers.clear
        theMatchers
    }
}